#define _GNU_SOURCE
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <elf.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <unistd.h>
#include <fcntl.h>
#include <setjmp.h>
#include <signal.h>
#include "kexec.h"
#include "crashdump.h"
#include "kexec-syscall.h"
#include "config.h"
#include "kexec-xen.h"

struct crash_note_info {
	unsigned long base;
	unsigned long length;
};

static int xen_phys_cpus;
static struct crash_note_info *xen_phys_notes;

#define XEN_MEM0_DIR "/sys/devices/system/xen_memory/xen_memory0"
#define XEN_MEM0_TARGET XEN_MEM0_DIR "/target_kb"
#define XEN_MEM0_LOW XEN_MEM0_DIR "/info/low_kb"
#define XEN_MEM0_HIGH XEN_MEM0_DIR "/info/high_kb"

/* based on code from xen-detect.c */
static int is_dom0;
#if defined(__i386__) || defined(__x86_64__)
enum {
	XEN_PV = 1,
	XEN_HVM = 2,
	XEN_NONE = 3,
};
static int guest_type;
static jmp_buf xen_sigill_jmp;
void xen_sigill_handler(int sig)
{
	longjmp(xen_sigill_jmp, 1);
}

static void xen_cpuid(uint32_t idx, uint32_t *regs, int pv_context)
{
#ifdef __i386__
    /* Use the stack to avoid reg constraint failures with some gcc flags */
    asm volatile (
        "push %%eax; push %%ebx; push %%ecx; push %%edx\n\t"
        "test %1,%1 ; jz 1f ; ud2a ; .ascii \"xen\" ; 1: cpuid\n\t"
        "mov %%eax,(%2); mov %%ebx,4(%2)\n\t"
        "mov %%ecx,8(%2); mov %%edx,12(%2)\n\t"
        "pop %%edx; pop %%ecx; pop %%ebx; pop %%eax\n\t"
        : : "a" (idx), "c" (pv_context), "S" (regs) : "memory" );
#else
    asm volatile (
        "test %5,%5 ; jz 1f ; ud2a ; .ascii \"xen\" ; 1: cpuid\n\t"
        : "=a" (regs[0]), "=b" (regs[1]), "=c" (regs[2]), "=d" (regs[3])
        : "0" (idx), "1" (pv_context), "2" (0) );
#endif
}

static int check_for_xen(int pv_context)
{
	uint32_t regs[4];
	char signature[13];
	uint32_t base;

	for (base = 0x40000000; base < 0x40010000; base += 0x100)
	{
		xen_cpuid(base, regs, pv_context);

		*(uint32_t *)(signature + 0) = regs[1];
		*(uint32_t *)(signature + 4) = regs[2];
		*(uint32_t *)(signature + 8) = regs[3];
		signature[12] = '\0';

		if (strcmp("XenVMMXenVMM", signature) == 0 && regs[0] >= (base + 2))
			goto found;
	}

	return 0;

found:
	xen_cpuid(base + 1, regs, pv_context);
	return regs[0];
}

static void xen_detect_guest_type(void)
{
	struct sigaction act, oldact;

	guest_type = XEN_NONE;
	if (check_for_xen(0)) {
		guest_type = XEN_HVM;
		return;
	}

	if (setjmp(xen_sigill_jmp))
		return;

	memset(&act, 0, sizeof(act));
	act.sa_handler = xen_sigill_handler;
	sigemptyset (&act.sa_mask);
	if (sigaction(SIGILL, &act, &oldact))
		return;
	if (check_for_xen(1))
		guest_type = XEN_PV;
	sigaction(SIGILL, &oldact, NULL);
	return;
}

static int xen_detect_pv_guest(void)
{
	if (!guest_type)
		xen_detect_guest_type();

	return guest_type == XEN_PV ? 1 : -1;
}

static int do_balloon_up(void)
{
	char line[123];
	FILE *f;
	int done = 0, seen_lo, seen_hi;
	long long lo, hi, prev_lo = 0, prev_hi = 0;

	if (!guest_type)
		xen_detect_guest_type();

	if (guest_type != XEN_HVM)
		return 0;

	/* Nothing to do if no balloon driver */
	f = fopen(XEN_MEM0_TARGET, "w");
	if (!f)
		return 0;

	/* Balloon up to maximum, the guest can not exceed its max_memkb */
	printf("Ballooning up in PVonHVM guest.\n");
	snprintf(line, sizeof(line), "%llu", -1LL);
	fwrite(line, strlen(line), 1, f);
	fclose(f);

	do {
		struct timeval timeout = {.tv_usec = 654321, };
		seen_lo = seen_hi = 0;
		lo = hi = -1;

		/* Wait for balloon driver to reach maximum */
		if (select(0, NULL, NULL, NULL, &timeout) < 0) {
			perror("select");
			break;
		}

		/* Check ballooned low mem */
		f = fopen(XEN_MEM0_LOW, "r");
		if (!f)
			break;
		if (fscanf(f, "%lld", &lo) == 1)
			seen_lo = 1;
		fclose(f);

		/* Check ballooned high mem */
		f = fopen(XEN_MEM0_HIGH, "r");
		if (!f)
			break;
		if (fscanf(f, "%lld", &hi) == 1)
			seen_hi = 1;
		fclose(f);

		/* Print progress if current values changed */
		if ((seen_lo || seen_hi) && (hi || lo) && (lo != prev_lo || hi != prev_hi)) {
			printf("h: %lld, l: %lld\n", hi, lo);
			if (seen_lo && lo != prev_lo)
				prev_lo = lo;
			if (seen_hi && hi != prev_hi)
				prev_hi = hi;
		}

		/* Exit loop if nothing is ballooned anymore */
		if (seen_lo && seen_hi && hi == 0 && lo == 0)
			done = 1;

	} while (!done);

	printf("%s.\n", done ? "Done" : "Not done");
	return !done;
}

#else
static int xen_detect_pv_guest(void)
{
	return 1;
}

static int do_balloon_up(void)
{
	return 0;
}
#endif

/*
 * Return 1 if its a PV guest.
 * This includes dom0, which is the only PV guest where kexec/kdump works.
 * HVM guests have to be handled as native hardware.
 */
int xen_present(void)
{
	if (!is_dom0) {
		if (access("/proc/xen", F_OK) == 0)
			is_dom0 = xen_detect_pv_guest();
		else
			is_dom0 = -1;
	}
	return is_dom0 > 0;
}

int xen_balloon_up(void)
{
	return do_balloon_up();
}

unsigned long xen_architecture(struct crash_elf_info *elf_info)
{
	unsigned long machine = elf_info->machine;
#ifdef HAVE_LIBXENCTRL
	int rc;
	xen_capabilities_info_t capabilities;
	xc_interface *xc;

	if (!xen_present())
		goto out;

	memset(capabilities, '0', XEN_CAPABILITIES_INFO_LEN);

	xc = xc_interface_open(NULL, NULL, 0);
	if ( !xc ) {
		fprintf(stderr, "failed to open xen control interface.\n");
		goto out;
	}

	rc = xc_version(xc, XENVER_capabilities, &capabilities[0]);
	if ( rc == -1 ) {
		fprintf(stderr, "failed to make Xen version hypercall.\n");
		goto out_close;
	}

	if (strstr(capabilities, "xen-3.0-x86_64"))
		machine = EM_X86_64;
        else if (strstr(capabilities, "xen-3.0-x86_32"))
		machine = EM_386;

 out_close:
	xc_interface_close(xc);

 out:
#endif
	return machine;
}

#ifdef HAVE_LIBXENCTRL
int get_xen_vmcoreinfo(uint64_t *addr, uint64_t *len)
{
	xc_interface *xc;
	int ret = 0;

	xc = xc_interface_open(NULL, NULL, 0);
	if (!xc) {
	        fprintf(stderr, "failed to open xen control interface.\n");
	        return -1;
	}

	ret = xc_kexec_get_range(xc, KEXEC_RANGE_MA_VMCOREINFO, 0, len, addr);

	xc_interface_close(xc);

	if (ret < 0)
	        return -1;
	return 0;
}

int xen_get_nr_phys_cpus(void)
{
	xc_interface *xc;
	int max_cpus;
	int cpu = -1;

	if (xen_phys_cpus)
		return xen_phys_cpus;

	xc = xc_interface_open(NULL, NULL, 0);
	if (!xc) {
		fprintf(stderr, "failed to open xen control interface.\n");
		return -1;
	}

	max_cpus = xc_get_max_cpus(xc);
	if (max_cpus <= 0)
		goto out;

	xen_phys_notes = calloc(max_cpus, sizeof(*xen_phys_notes));
	if (xen_phys_notes == NULL)
		goto out;

	for (cpu = 0; cpu < max_cpus; cpu++) {
		uint64_t size, start;
		int ret;

		ret = xc_kexec_get_range(xc, KEXEC_RANGE_MA_CPU, cpu, &size, &start);
		if (ret < 0)
			break;

		xen_phys_notes[cpu].base = start;
		xen_phys_notes[cpu].length = size;
	}

	xen_phys_cpus = cpu;

out:
	xc_interface_close(xc);
	return cpu;
}
#else
int get_xen_vmcoreinfo(uint64_t *addr, uint64_t *len)
{
	return -1;
}

int xen_get_nr_phys_cpus(void)
{
	return -1;
}
#endif


int xen_get_note(int cpu, uint64_t *addr, uint64_t *len)
{
	struct crash_note_info *note;

	if (xen_phys_cpus <= 0)
		return -1;

	note = xen_phys_notes + cpu;

	*addr = note->base;
	*len = note->length;

	return 0;
}

#ifdef HAVE_LIBXENCTRL
int xen_get_crashkernel_region(uint64_t *start, uint64_t *end)
{
	uint64_t size;
	xc_interface *xc;
	int rc = -1;

	xc = xc_interface_open(NULL, NULL, 0);
	if (!xc) {
		fprintf(stderr, "failed to open xen control interface.\n");
		goto out;
	}

	rc = xc_kexec_get_range(xc, KEXEC_RANGE_MA_CRASH, 0, &size, start);
	if (rc < 0) {
		fprintf(stderr, "failed to get crash region from hypervisor.\n");
		goto out_close;
	}

	*end = *start + size - 1;

out_close:
	xc_interface_close(xc);

out:
	return rc;
}
#else
int xen_get_crashkernel_region(uint64_t *start, uint64_t *end)
{
	return -1;
}
#endif
